"""A custom hatch build hook for ipykernel."""

import shutil
import sys
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomHook(BuildHookInterface):
    """The IPykernel build hook."""

    def initialize(self, version, build_data):
        """Initialize the hook."""
        here = Path(__file__).parent.resolve()
        sys.path.insert(0, str(here))
        from ipykernel.kernelspec import make_ipkernel_cmd, write_kernel_spec

        overrides = {}

        # When building a standard wheel, the executable specified in the kernelspec is simply 'python'.
        if version == "standard":
            overrides["metadata"] = dict(debugger=True)
            argv = make_ipkernel_cmd(executable="python")

        # When installing an editable wheel, the full `sys.executable` can be used.
        else:
            argv = make_ipkernel_cmd()

        overrides["argv"] = argv

        dest = Path(here) / "data_kernelspec"
        if Path(dest).exists():
            shutil.rmtree(dest)

        write_kernel_spec(dest, overrides=overrides)



import os
import sys
import warnings
from unittest import mock

import pytest

from ipykernel.parentpoller import ParentPollerUnix, ParentPollerWindows


@pytest.mark.skipif(os.name == "nt", reason="only works on posix")
def test_parent_poller_unix_to_pid1():
    poller = ParentPollerUnix()
    with mock.patch("os.getppid", lambda: 1):  # noqa: PT008

        def exit_mock(*args):
            sys.exit(1)

        with mock.patch("os._exit", exit_mock), pytest.raises(SystemExit):
            poller.run()

    def mock_getppid():
        msg = "hi"
        raise ValueError(msg)

    with mock.patch("os.getppid", mock_getppid), pytest.raises(ValueError):  # noqa: PT011
        poller.run()


@pytest.mark.skipif(os.name == "nt", reason="only works on posix")
def test_parent_poller_unix_reparent_not_pid1():
    parent_pid = 221
    parent_pids = iter([parent_pid, parent_pid - 1])

    poller = ParentPollerUnix(parent_pid=parent_pid)

    with mock.patch("os.getppid", lambda: next(parent_pids)):  # noqa: PT008

        def exit_mock(*args):
            sys.exit(1)

        with mock.patch("os._exit", exit_mock), pytest.raises(SystemExit):
            poller.run()


@pytest.mark.skipif(os.name != "nt", reason="only works on windows")
def test_parent_poller_windows():
    poller = ParentPollerWindows(interrupt_handle=1)

    def mock_wait(*args, **kwargs):
        return -1

    with mock.patch("ctypes.windll.kernel32.WaitForMultipleObjects", mock_wait):  # noqa
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            poller.run()



"""An in-process terminal example."""

import os
import sys

import tornado
from jupyter_console.ptshell import ZMQTerminalInteractiveShell

from ipykernel.inprocess.manager import InProcessKernelManager


def print_process_id():
    """Print the process id."""
    print("Process ID is:", os.getpid())


def init_asyncio_patch():
    """set default asyncio policy to be compatible with tornado
    Tornado 6 (at least) is not compatible with the default
    asyncio implementation on Windows
    Pick the older SelectorEventLoopPolicy on Windows
    if the known-incompatible default policy is in use.
    do this as early as possible to make it a low priority and overridable
    ref: https://github.com/tornadoweb/tornado/issues/2608
    FIXME: if/when tornado supports the defaults in asyncio,
           remove and bump tornado requirement for py38
    """
    if (
        sys.platform.startswith("win")
        and sys.version_info >= (3, 8)
        and tornado.version_info < (6, 1)
    ):
        import asyncio

        try:
            from asyncio import WindowsProactorEventLoopPolicy, WindowsSelectorEventLoopPolicy
        except ImportError:
            pass
            # not affected
        else:
            if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                # fallback to the pre-3.8 default of Selector
                asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def main():
    """The main function."""
    print_process_id()

    # Create an in-process kernel
    # >>> print_process_id()
    # will print the same process ID as the main process
    init_asyncio_patch()
    kernel_manager = InProcessKernelManager()
    kernel_manager.start_kernel()
    kernel = kernel_manager.kernel
    kernel.gui = "qt4"
    kernel.shell.push({"foo": 43, "print_process_id": print_process_id})
    client = kernel_manager.client()
    client.start_channels()

    shell = ZMQTerminalInteractiveShell(manager=kernel_manager, client=client)
    shell.mainloop()


if __name__ == "__main__":
    main()



import os
import sys
import time

import pytest
from jupyter_client.blocking.client import BlockingKernelClient

from .test_eventloop import qt_guis_avail
from .utils import assemble_output

# these tests don't seem to work with xvfb yet
# these tests seem to be a problem on CI in general
pytestmark = pytest.mark.skipif(
    bool(os.getenv("CI")),
    reason="tests not working yet reliably on CI",
)

guis = []
if not sys.platform.startswith("tk"):
    guis.append("tk")
if qt_guis_avail:
    guis.append("qt")
if sys.platform == "darwin":
    guis.append("osx")

backends = {
    "tk": "tkagg",
    "qt": "qtagg",
    "osx": "macosx",
}


def execute(
    kc: BlockingKernelClient,
    code: str,
    timeout=120,
):
    msg_id = kc.execute(code)
    stdout, stderr = assemble_output(kc.get_iopub_msg, timeout=timeout, parent_msg_id=msg_id)
    assert not stderr.strip()
    return stdout.strip(), stderr.strip()


@pytest.mark.parametrize("gui", guis)
@pytest.mark.timeout(300)
def test_matplotlib_gui(kc, gui):
    """Make sure matplotlib activates and its eventloop runs while the kernel is also responsive"""
    pytest.importorskip("matplotlib", reason="this test requires matplotlib")
    stdout, stderr = execute(kc, f"%matplotlib {gui}")
    assert not stderr
    # debug: show output from invoking the matplotlib magic
    print(stdout)
    execute(
        kc,
        """
    from concurrent.futures import Future
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    """,
    )
    stdout, _ = execute(kc, "print(mpl.get_backend())")
    assert stdout == backends[gui]
    execute(
        kc,
        """
fig, ax = plt.subplots()
timer = fig.canvas.new_timer(interval=10)
f = Future()

call_count = 0
def add_call():
    global call_count
    call_count += 1
    if not f.done():
        f.set_result(None)

timer.add_callback(add_call)
timer.start()
""",
    )
    # wait for the first call (up to 60 seconds)
    deadline = time.monotonic() + 60
    done = False
    while time.monotonic() <= deadline:
        stdout, _ = execute(kc, "print(f.done())")
        if stdout.strip() == "True":
            done = True
            break
        if stdout == "False":
            time.sleep(0.1)
        else:
            pytest.fail(f"Unexpected output {stdout}")
    if not done:
        pytest.fail("future never finished...")

    time.sleep(0.25)
    stdout, _ = execute(kc, "print(call_count)")
    call_count = int(stdout)
    assert call_count > 0
    time.sleep(0.25)
    stdout, _ = execute(kc, "timer.stop()\nprint(call_count)")
    call_count_2 = int(stdout)
    assert call_count_2 > call_count
    stdout, _ = execute(kc, "print(call_count)")
    call_count_3 = int(stdout)
    assert call_count_3 <= call_count_2 + 5



"""Manager of subshells in a kernel."""

from __future__ import annotations

import asyncio
import json
import typing as t
import uuid
from functools import partial
from threading import Lock, current_thread

import zmq
from tornado.ioloop import IOLoop

from .socket_pair import SocketPair
from .subshell import SubshellThread
from .thread import SHELL_CHANNEL_THREAD_NAME
from .utils import _async_in_context


class SubshellManager:
    """A manager of subshells.

    Controls the lifetimes of subshell threads and their associated ZMQ sockets and
    streams. Runs mostly in the shell channel thread.

    Care needed with threadsafe access here.  All write access to the cache occurs in
    the shell channel thread so there is only ever one write access at any one time.
    Reading of cache information can be performed by other threads, so all reads are
    protected by a lock so that they are atomic.

    Sending reply messages via the shell_socket is wrapped by another lock to protect
    against multiple subshells attempting to send at the same time.

    .. versionadded:: 7
    """

    def __init__(
        self,
        context: zmq.Context[t.Any],
        shell_channel_io_loop: IOLoop,
        shell_socket: zmq.Socket[t.Any],
    ):
        """Initialize the subshell manager."""
        self._parent_thread = current_thread()

        self._context: zmq.Context[t.Any] = context
        self._shell_channel_io_loop = shell_channel_io_loop
        self._shell_socket = shell_socket
        self._cache: dict[str, SubshellThread] = {}
        self._lock_cache = Lock()  # Sync lock across threads when accessing cache.

        # Inproc socket pair for communication from control thread to shell channel thread,
        # such as for create_subshell_request messages. Reply messages are returned straight away.
        self.control_to_shell_channel = SocketPair(self._context, "control")
        self.control_to_shell_channel.on_recv(
            self._shell_channel_io_loop, self._process_control_request, copy=True
        )

        # Inproc socket pair for communication from shell channel thread to main thread,
        # such as for execute_request messages.
        self._shell_channel_to_main = SocketPair(self._context, "main")

        # Inproc socket pair for communication from main thread to shell channel thread.
        # such as for execute_reply messages.
        self._main_to_shell_channel = SocketPair(self._context, "main-reverse")
        self._main_to_shell_channel.on_recv(
            self._shell_channel_io_loop, self._send_on_shell_channel
        )

    def close(self) -> None:
        """Stop all subshells and close all resources."""
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME
        with self._lock_cache:
            while True:
                try:
                    _, subshell_thread = self._cache.popitem()
                except KeyError:
                    break
                self._stop_subshell(subshell_thread)

        self.control_to_shell_channel.close()
        self._main_to_shell_channel.close()
        self._shell_channel_to_main.close()

    def get_shell_channel_to_subshell_pair(self, subshell_id: str | None) -> SocketPair:
        """Return the inproc socket pair used to send messages from the shell channel
        to a particular subshell or main shell."""
        if subshell_id is None:
            return self._shell_channel_to_main
        with self._lock_cache:
            return self._cache[subshell_id].shell_channel_to_subshell

    def get_subshell_to_shell_channel_socket(self, subshell_id: str | None) -> zmq.Socket[t.Any]:
        """Return the socket used by a particular subshell or main shell to send
        messages to the shell channel.
        """
        if subshell_id is None:
            return self._main_to_shell_channel.from_socket
        with self._lock_cache:
            return self._cache[subshell_id].subshell_to_shell_channel.from_socket

    def get_shell_channel_to_subshell_socket(self, subshell_id: str | None) -> zmq.Socket[t.Any]:
        """Return the socket used by the shell channel to send messages to a particular
        subshell or main shell.
        """
        return self.get_shell_channel_to_subshell_pair(subshell_id).from_socket

    def get_subshell_aborting(self, subshell_id: str) -> bool:
        """Get the boolean aborting flag of the specified subshell."""
        with self._lock_cache:
            return self._cache[subshell_id].aborting

    def get_subshell_asyncio_lock(self, subshell_id: str) -> asyncio.Lock:
        """Return the asyncio lock belonging to the specified subshell."""
        with self._lock_cache:
            return self._cache[subshell_id].asyncio_lock

    def list_subshell(self) -> list[str]:
        """Return list of current subshell ids.

        Can be called by any subshell using %subshell magic.
        """
        with self._lock_cache:
            return list(self._cache)

    def set_on_recv_callback(self, on_recv_callback):
        """Set the callback used by the main shell and all subshells to receive
        messages sent from the shell channel thread.
        """
        assert current_thread() == self._parent_thread
        self._on_recv_callback = on_recv_callback
        self._shell_channel_to_main.on_recv(
            IOLoop.current(), _async_in_context(partial(on_recv_callback, None))
        )

    def set_subshell_aborting(self, subshell_id: str, aborting: bool) -> None:
        """Set the aborting flag of the specified subshell."""
        with self._lock_cache:
            self._cache[subshell_id].aborting = aborting

    def subshell_id_from_thread_id(self, thread_id: int) -> str | None:
        """Return subshell_id of the specified thread_id.

        Raises RuntimeError if thread_id is not the main shell or a subshell.

        Only used by %subshell magic so does not have to be fast/cached.
        """
        with self._lock_cache:
            if thread_id == self._parent_thread.ident:
                return None
            for id, subshell in self._cache.items():
                if subshell.ident == thread_id:
                    return id
            msg = f"Thread id {thread_id!r} does not correspond to a subshell of this kernel"
            raise RuntimeError(msg)

    def _create_subshell(self) -> str:
        """Create and start a new subshell thread."""
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME

        subshell_id = str(uuid.uuid4())
        subshell_thread = SubshellThread(subshell_id, self._context)

        with self._lock_cache:
            assert subshell_id not in self._cache
            self._cache[subshell_id] = subshell_thread

        subshell_thread.shell_channel_to_subshell.on_recv(
            subshell_thread.io_loop,
            _async_in_context(partial(self._on_recv_callback, subshell_id)),
        )

        subshell_thread.subshell_to_shell_channel.on_recv(
            self._shell_channel_io_loop, self._send_on_shell_channel
        )

        subshell_thread.start()
        return subshell_id

    def _delete_subshell(self, subshell_id: str) -> None:
        """Delete subshell identified by subshell_id.

        Raises key error if subshell_id not in cache.
        """
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME

        with self._lock_cache:
            subshell_threwad = self._cache.pop(subshell_id)

        self._stop_subshell(subshell_threwad)

    def _process_control_request(
        self,
        request: list[t.Any],
    ) -> None:
        """Process a control request message received on the control inproc
        socket and return the reply.  Runs in the shell channel thread.
        """
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME

        try:
            decoded = json.loads(request[0])
            type = decoded["type"]
            reply: dict[str, t.Any] = {"status": "ok"}

            if type == "create":
                reply["subshell_id"] = self._create_subshell()
            elif type == "delete":
                subshell_id = decoded["subshell_id"]
                self._delete_subshell(subshell_id)
            elif type == "list":
                reply["subshell_id"] = self.list_subshell()
            else:
                msg = f"Unrecognised message type {type!r}"
                raise RuntimeError(msg)
        except BaseException as err:
            reply = {
                "status": "error",
                "evalue": str(err),
            }

        # Return the reply to the control thread.
        self.control_to_shell_channel.to_socket.send_json(reply)

    def _send_on_shell_channel(self, msg) -> None:
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME
        self._shell_socket.send_multipart(msg)

    def _stop_subshell(self, subshell_thread: SubshellThread) -> None:
        """Stop a subshell thread and close all of its resources."""
        assert current_thread().name == SHELL_CHANNEL_THREAD_NAME

        if subshell_thread.is_alive():
            subshell_thread.stop()
            subshell_thread.join()



"""An internal ipykernel example."""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import sys

from IPython.lib.kernel import connect_qtconsole

from ipykernel.kernelapp import IPKernelApp


# -----------------------------------------------------------------------------
# Functions and classes
# -----------------------------------------------------------------------------
def mpl_kernel(gui):
    """Launch and return an IPython kernel with matplotlib support for the desired gui"""
    kernel = IPKernelApp.instance()
    kernel.initialize(
        [
            "python",
            "--matplotlib=%s" % gui,
            #'--log-level=10'
        ]
    )
    return kernel


class InternalIPKernel:
    """An internal ipykernel class."""

    def init_ipkernel(self, backend):
        """Start IPython kernel with GUI event loop and mpl support."""
        self.ipkernel = mpl_kernel(backend)
        # To create and track active qt consoles
        self.consoles = []

        # This application will also act on the shell user namespace
        self.namespace = self.ipkernel.shell.user_ns

        # Example: a variable that will be seen by the user in the shell, and
        # that the GUI modifies (the 'Counter++' button increments it):
        self.namespace["app_counter"] = 0
        # self.namespace['ipkernel'] = self.ipkernel  # dbg

    def print_namespace(self, evt=None):
        """Print the namespace."""
        print("\n***Variables in User namespace***")
        for k, v in self.namespace.items():
            if not k.startswith("_"):
                print(f"{k} -> {v!r}")
        sys.stdout.flush()

    def new_qt_console(self, evt=None):
        """start a new qtconsole connected to our kernel"""
        return connect_qtconsole(self.ipkernel.abs_connection_file, profile=self.ipkernel.profile)

    def count(self, evt=None):
        """Get the app counter value."""
        self.namespace["app_counter"] += 1

    def cleanup_consoles(self, evt=None):
        """Clean up the consoles."""
        for c in self.consoles:
            c.kill()



import sys

import pytest

from .utils import TIMEOUT, get_replies, get_reply, new_kernel

seq = 0

# Tests support debugpy not being installed, in which case the tests don't do anything useful
# functionally as the debug message replies are usually empty dictionaries, but they confirm that
# ipykernel doesn't block, or segfault, or raise an exception.
try:
    import debugpy
except ImportError:
    debugpy = None


def prepare_debug_request(kernel, command, arguments=None):
    """Prepare a debug request but do not send it."""
    global seq
    seq += 1

    msg = kernel.session.msg(
        "debug_request",
        {
            "type": "request",
            "seq": seq,
            "command": command,
            "arguments": arguments or {},
        },
    )
    return msg


def wait_for_debug_request(kernel, command, arguments=None, full_reply=False):
    """Carry out a debug request and return the reply content.

    It does not check if the request was successful.
    """
    msg = prepare_debug_request(kernel, command, arguments)
    kernel.control_channel.send(msg)
    reply = get_reply(kernel, msg["header"]["msg_id"], channel="control")
    return reply if full_reply else reply["content"]


@pytest.fixture()
def kernel():
    with new_kernel() as kc:
        yield kc


@pytest.fixture()
def kernel_with_debug(kernel):
    # Initialize
    wait_for_debug_request(
        kernel,
        "initialize",
        {
            "clientID": "test-client",
            "clientName": "testClient",
            "adapterID": "",
            "pathFormat": "path",
            "linesStartAt1": True,
            "columnsStartAt1": True,
            "supportsVariableType": True,
            "supportsVariablePaging": True,
            "supportsRunInTerminalRequest": True,
            "locale": "en",
        },
    )

    # Attach
    wait_for_debug_request(kernel, "attach")

    try:
        yield kernel
    finally:
        # Detach
        wait_for_debug_request(kernel, "disconnect", {"restart": False, "terminateDebuggee": True})


def test_debug_initialize(kernel):
    reply = wait_for_debug_request(
        kernel,
        "initialize",
        {
            "clientID": "test-client",
            "clientName": "testClient",
            "adapterID": "",
            "pathFormat": "path",
            "linesStartAt1": True,
            "columnsStartAt1": True,
            "supportsVariableType": True,
            "supportsVariablePaging": True,
            "supportsRunInTerminalRequest": True,
            "locale": "en",
        },
    )
    if debugpy:
        assert reply["success"]
    else:
        assert reply == {}


def test_supported_features(kernel_with_debug):
    kernel_with_debug.kernel_info()
    reply = kernel_with_debug.get_shell_msg(timeout=TIMEOUT)
    supported_features = reply["content"]["supported_features"]

    if debugpy:
        assert "debugger" in supported_features
    else:
        assert "debugger" not in supported_features


def test_attach_debug(kernel_with_debug):
    reply = wait_for_debug_request(
        kernel_with_debug, "evaluate", {"expression": "'a' + 'b'", "context": "repl"}
    )
    if debugpy:
        assert reply["success"]
        assert reply["body"]["result"] == ""
    else:
        assert reply == {}


def test_set_breakpoints(kernel_with_debug):
    code = """def f(a, b):
    c = a + b
    return c

f(2, 3)"""

    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "non-existent path"

    reply = wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 2}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )
    if debugpy:
        assert reply["success"]
        assert len(reply["body"]["breakpoints"]) == 1
        assert reply["body"]["breakpoints"][0]["verified"]
        assert reply["body"]["breakpoints"][0]["source"]["path"] == source
    else:
        assert reply == {}

    r = wait_for_debug_request(kernel_with_debug, "debugInfo")

    def func(b):
        return b["source"]

    if debugpy:
        assert source in map(func, r["body"]["breakpoints"])
    else:
        assert r == {}

    r = wait_for_debug_request(kernel_with_debug, "configurationDone")
    if debugpy:
        assert r["success"]
    else:
        assert r == {}


def test_stop_on_breakpoint(kernel_with_debug):
    code = """def f(a, b):
    c = a + b
    return c

f(2, 3)"""

    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "some path"

    wait_for_debug_request(kernel_with_debug, "debugInfo")

    wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 2}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )

    wait_for_debug_request(kernel_with_debug, "configurationDone", full_reply=True)

    kernel_with_debug.execute(code)

    if not debugpy:
        # Cannot stop on breakpoint if debugpy not installed
        return

    # Wait for stop on breakpoint
    msg: dict = {"msg_type": "", "content": {}}
    while msg.get("msg_type") != "debug_event" or msg["content"].get("event") != "stopped":
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)

    assert msg["content"]["body"]["reason"] == "breakpoint"


def test_breakpoint_in_cell_with_leading_empty_lines(kernel_with_debug):
    code = """
def f(a, b):
    c = a + b
    return c

f(2, 3)"""

    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "some path"

    wait_for_debug_request(kernel_with_debug, "debugInfo")

    wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 6}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )

    wait_for_debug_request(kernel_with_debug, "configurationDone", full_reply=True)

    kernel_with_debug.execute(code)

    if not debugpy:
        # Cannot stop on breakpoint if debugpy not installed
        return

    # Wait for stop on breakpoint
    msg: dict = {"msg_type": "", "content": {}}
    while msg.get("msg_type") != "debug_event" or msg["content"].get("event") != "stopped":
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)

    assert msg["content"]["body"]["reason"] == "breakpoint"


def test_rich_inspect_not_at_breakpoint(kernel_with_debug):
    var_name = "text"
    value = "Hello the world"
    code = f"""{var_name}='{value}'
print({var_name})
"""

    msg_id = kernel_with_debug.execute(code)
    get_reply(kernel_with_debug, msg_id)

    r = wait_for_debug_request(kernel_with_debug, "inspectVariables")

    def func(v):
        return v["name"]

    if debugpy:
        assert var_name in list(map(func, r["body"]["variables"]))
    else:
        assert r == {}

    reply = wait_for_debug_request(
        kernel_with_debug,
        "richInspectVariables",
        {"variableName": var_name},
    )

    if debugpy:
        assert reply["body"]["data"] == {"text/plain": f"'{value}'"}
    else:
        assert reply == {}


def test_rich_inspect_at_breakpoint(kernel_with_debug):
    code = """def f(a, b):
    c = a + b
    return c

f(2, 3)"""

    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "some path"

    wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 2}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )

    r = wait_for_debug_request(kernel_with_debug, "debugInfo")

    r = wait_for_debug_request(kernel_with_debug, "configurationDone")

    kernel_with_debug.execute(code)

    if not debugpy:
        # Cannot stop on breakpoint if debugpy not installed
        return

    # Wait for stop on breakpoint
    msg: dict = {"msg_type": "", "content": {}}
    while msg.get("msg_type") != "debug_event" or msg["content"].get("event") != "stopped":
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)

    stacks = wait_for_debug_request(kernel_with_debug, "stackTrace", {"threadId": 1})["body"][
        "stackFrames"
    ]

    scopes = wait_for_debug_request(kernel_with_debug, "scopes", {"frameId": stacks[0]["id"]})[
        "body"
    ]["scopes"]

    locals_ = wait_for_debug_request(
        kernel_with_debug,
        "variables",
        {
            "variablesReference": next(filter(lambda s: s["name"] == "Locals", scopes))[
                "variablesReference"
            ]
        },
    )["body"]["variables"]

    reply = wait_for_debug_request(
        kernel_with_debug,
        "richInspectVariables",
        {"variableName": locals_[0]["name"], "frameId": stacks[0]["id"]},
    )

    assert reply["body"]["data"] == {"text/plain": locals_[0]["value"]}


def test_convert_to_long_pathname():
    if sys.platform == "win32":
        from ipykernel.compiler import _convert_to_long_pathname

        _convert_to_long_pathname(__file__)


def test_copy_to_globals(kernel_with_debug):
    local_var_name = "var"
    global_var_name = "var_copy"
    code = f"""from IPython.core.display import HTML
def my_test():
    {local_var_name} = HTML('<p>test content</p>')
    pass
a = 2
my_test()"""

    # Init debugger and set breakpoint
    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "some path"

    wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 4}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )

    wait_for_debug_request(kernel_with_debug, "debugInfo")

    wait_for_debug_request(kernel_with_debug, "configurationDone")

    # Execute code
    kernel_with_debug.execute(code)

    if not debugpy:
        # Cannot stop on breakpoint if debugpy not installed
        return

    # Wait for stop on breakpoint
    msg: dict = {"msg_type": "", "content": {}}
    while msg.get("msg_type") != "debug_event" or msg["content"].get("event") != "stopped":
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)

    stacks = wait_for_debug_request(kernel_with_debug, "stackTrace", {"threadId": 1})["body"][
        "stackFrames"
    ]

    # Get local frame id
    frame_id = stacks[0]["id"]

    # Copy the variable
    wait_for_debug_request(
        kernel_with_debug,
        "copyToGlobals",
        {
            "srcVariableName": local_var_name,
            "dstVariableName": global_var_name,
            "srcFrameId": frame_id,
        },
    )

    # Get the scopes
    scopes = wait_for_debug_request(kernel_with_debug, "scopes", {"frameId": frame_id})["body"][
        "scopes"
    ]

    # Get the local variable
    locals_ = wait_for_debug_request(
        kernel_with_debug,
        "variables",
        {
            "variablesReference": next(filter(lambda s: s["name"] == "Locals", scopes))[
                "variablesReference"
            ]
        },
    )["body"]["variables"]

    local_var = None
    for variable in locals_:
        if local_var_name in variable["evaluateName"]:
            local_var = variable
    assert local_var is not None

    # Get the global variable (copy of the local variable)
    globals_ = wait_for_debug_request(
        kernel_with_debug,
        "variables",
        {
            "variablesReference": next(filter(lambda s: s["name"] == "Globals", scopes))[
                "variablesReference"
            ]
        },
    )["body"]["variables"]

    global_var = None
    for variable in globals_:
        if global_var_name in variable["evaluateName"]:
            global_var = variable
    assert global_var is not None

    # Compare local and global variable
    assert global_var["value"] == local_var["value"] and global_var["type"] == local_var["type"]  # noqa: PT018


def test_debug_requests_sequential(kernel_with_debug):
    # Issue https://github.com/ipython/ipykernel/issues/1412
    # Control channel requests should be executed sequentially not concurrently.
    code = """def f(a, b):
    c = a + b
    return c

f(2, 3)"""

    r = wait_for_debug_request(kernel_with_debug, "dumpCell", {"code": code})
    if debugpy:
        source = r["body"]["sourcePath"]
    else:
        assert r == {}
        source = "some path"

    wait_for_debug_request(
        kernel_with_debug,
        "setBreakpoints",
        {
            "breakpoints": [{"line": 2}],
            "source": {"path": source},
            "sourceModified": False,
        },
    )

    wait_for_debug_request(kernel_with_debug, "debugInfo")
    wait_for_debug_request(kernel_with_debug, "configurationDone")
    kernel_with_debug.execute(code)

    if not debugpy:
        # Cannot stop on breakpoint if debugpy not installed
        return

    # Wait for stop on breakpoint
    msg: dict = {"msg_type": "", "content": {}}
    while msg.get("msg_type") != "debug_event" or msg["content"].get("event") != "stopped":
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)

    stacks = wait_for_debug_request(kernel_with_debug, "stackTrace", {"threadId": 1})["body"][
        "stackFrames"
    ]

    scopes = wait_for_debug_request(kernel_with_debug, "scopes", {"frameId": stacks[0]["id"]})[
        "body"
    ]["scopes"]

    # Get variablesReference for both Locals and Globals.
    locals_ref = next(filter(lambda s: s["name"] == "Locals", scopes))["variablesReference"]
    globals_ref = next(filter(lambda s: s["name"] == "Globals", scopes))["variablesReference"]

    msgs = []
    for ref in [locals_ref, globals_ref]:
        msgs.append(
            prepare_debug_request(kernel_with_debug, "variables", {"variablesReference": ref})
        )

    # Send messages in quick succession.
    for msg in msgs:
        kernel_with_debug.control_channel.send(msg)

    replies = get_replies(kernel_with_debug, [msg["msg_id"] for msg in msgs], channel="control")

    # Check debug variable returns are correct.
    locals = replies[0]["content"]
    assert locals["success"]
    variables = locals["body"]["variables"]
    var = next(filter(lambda v: v["name"] == "a", variables))
    assert var["type"] == "int"
    assert var["value"] == "2"
    var = next(filter(lambda v: v["name"] == "b", variables))
    assert var["type"] == "int"
    assert var["value"] == "3"

    globals = replies[1]["content"]
    assert globals["success"]
    variables = globals["body"]["variables"]

    names = [v["name"] for v in variables]
    assert "function variables" in names
    assert "special variables" in names

    # Check status iopub messages alternate between busy and idle.
    execution_states = []
    while len(execution_states) < 8:
        msg = kernel_with_debug.get_iopub_msg(timeout=TIMEOUT)
        if msg["msg_type"] == "status":
            execution_states.append(msg["content"]["execution_state"])
    assert execution_states.count("busy") == 4
    assert execution_states.count("idle") == 4
    assert execution_states == ["busy", "idle"] * 4



"""Pair of ZMQ inproc sockets used for communication between threads."""

from __future__ import annotations

from typing import Any

import zmq
from tornado.ioloop import IOLoop
from zmq.eventloop.zmqstream import ZMQStream


class SocketPair:
    """Pair of ZMQ inproc sockets for one-direction communication between 2 threads.

    One of the threads is always the shell_channel_thread, the other may be the control
    thread, main thread or a subshell thread.

    .. versionadded:: 7
    """

    from_socket: zmq.Socket[Any]
    to_socket: zmq.Socket[Any]
    to_stream: ZMQStream | None = None

    def __init__(self, context: zmq.Context[Any], name: str):
        """Initialize the inproc socker pair."""
        self.from_socket = context.socket(zmq.PAIR)
        self.to_socket = context.socket(zmq.PAIR)
        address = self._address(name)
        self.from_socket.bind(address)
        self.to_socket.connect(address)  # Or do I need to do this in another thread?

    def close(self):
        """Close the inproc socker pair."""
        self.from_socket.close()

        if self.to_stream is not None:
            self.to_stream.close()
        self.to_socket.close()

    def on_recv(self, io_loop: IOLoop, on_recv_callback, copy: bool = False):
        """Set the callback used when a message is received on the to stream."""
        # io_loop is that of the 'to' thread.
        if self.to_stream is None:
            self.to_stream = ZMQStream(self.to_socket, io_loop)
        self.to_stream.on_recv(on_recv_callback, copy=copy)

    def _address(self, name) -> str:
        """Return the address used for this inproc socket pair."""
        return f"inproc://subshell{name}"



"""A thread for a subshell."""

import asyncio
from typing import Any

import zmq

from .socket_pair import SocketPair
from .thread import BaseThread


class SubshellThread(BaseThread):
    """A thread for a subshell.

    .. versionadded:: 7
    """

    def __init__(
        self,
        subshell_id: str,
        context: zmq.Context[Any],
        **kwargs,
    ):
        """Initialize the thread."""
        super().__init__(name=f"subshell-{subshell_id}", **kwargs)

        self.shell_channel_to_subshell = SocketPair(context, subshell_id)
        self.subshell_to_shell_channel = SocketPair(context, subshell_id + "-reverse")

        # When aborting flag is set, execute_request messages to this subshell will be aborted.
        self.aborting = False

        self.asyncio_lock = asyncio.Lock()

    def run(self) -> None:
        """Run the thread."""
        try:
            super().run()
        finally:
            self.shell_channel_to_subshell.close()
            self.subshell_to_shell_channel.close()



"""Utilities"""

from __future__ import annotations

import asyncio
import sys
import typing as t
from collections.abc import Mapping
from contextvars import copy_context
from functools import partial, wraps

if t.TYPE_CHECKING:
    from collections.abc import Callable
    from contextvars import Context


class LazyDict(Mapping[str, t.Any]):
    """Lazy evaluated read-only dictionary.

    Initialised with a dictionary of key-value pairs where the values are either
    constants or callables. Callables are evaluated each time the respective item is
    read.
    """

    def __init__(self, dict):
        self._dict = dict

    def __getitem__(self, key):
        item = self._dict.get(key)
        return item() if callable(item) else item

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)


T = t.TypeVar("T")
U = t.TypeVar("U")
V = t.TypeVar("V")


def _async_in_context(
    f: Callable[..., t.Coroutine[T, U, V]], context: Context | None = None
) -> Callable[..., t.Coroutine[T, U, V]]:
    """
    Wrapper to run a coroutine in a persistent ContextVar Context.

    Backports asyncio.create_task(context=...) behavior from Python 3.11
    """
    if context is None:
        context = copy_context()

    if sys.version_info >= (3, 11):

        @wraps(f)
        async def run_in_context(*args, **kwargs):
            coro = f(*args, **kwargs)
            return await asyncio.create_task(coro, context=context)

        return run_in_context

    # don't need this backport when we require 3.11
    # context_holder so we have a modifiable container for later calls
    context_holder = [context]  # type: ignore[unreachable]

    async def preserve_context(f, *args, **kwargs):
        """call a coroutine, preserving the context after it is called"""
        try:
            return await f(*args, **kwargs)
        finally:
            # persist changes to the context for future calls
            context_holder[0] = copy_context()

    @wraps(f)
    async def run_in_context_pre311(*args, **kwargs):
        ctx = context_holder[0]
        return await ctx.run(partial(asyncio.create_task, preserve_context(f, *args, **kwargs)))

    return run_in_context_pre311



"""Test kernel subshells."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import platform
import time
from collections import Counter
from queue import Empty

import pytest
from jupyter_client.blocking.client import BlockingKernelClient

from .utils import (
    TIMEOUT,
    assemble_output,
    flush_channels,
    get_replies,
    get_reply,
    new_kernel,
    wait_for_idle,
)

# Helpers


def create_subshell_helper(kc: BlockingKernelClient):
    msg = kc.session.msg("create_subshell_request")
    kc.control_channel.send(msg)
    msg_id = msg["header"]["msg_id"]
    reply = get_reply(kc, msg_id, TIMEOUT, channel="control")
    return reply["content"]


def delete_subshell_helper(kc: BlockingKernelClient, subshell_id: str):
    msg = kc.session.msg("delete_subshell_request", {"subshell_id": subshell_id})
    kc.control_channel.send(msg)
    msg_id = msg["header"]["msg_id"]
    reply = get_reply(kc, msg_id, TIMEOUT, channel="control")
    return reply["content"]


def list_subshell_helper(kc: BlockingKernelClient):
    msg = kc.session.msg("list_subshell_request")
    kc.control_channel.send(msg)
    msg_id = msg["header"]["msg_id"]
    reply = get_reply(kc, msg_id, TIMEOUT, channel="control")
    return reply["content"]


def execute_request(
    kc: BlockingKernelClient, code: str, subshell_id: str | None, silent: bool = False
):
    msg = kc.session.msg("execute_request", {"code": code, "silent": silent})
    msg["header"]["subshell_id"] = subshell_id
    kc.shell_channel.send(msg)
    return msg


def execute_request_subshell_id(
    kc: BlockingKernelClient, code: str, subshell_id: str | None, terminator: str = "\n"
):
    msg = execute_request(kc, code, subshell_id)
    msg_id = msg["header"]["msg_id"]
    stdout, _ = assemble_output(kc.get_iopub_msg, None, msg_id)
    return stdout.strip()


def execute_thread_count(kc: BlockingKernelClient) -> int:
    code = "print(threading.active_count())"
    return int(execute_request_subshell_id(kc, code, None))


def execute_thread_ids(kc: BlockingKernelClient, subshell_id: str | None = None) -> tuple[str, str]:
    code = "print(threading.get_ident(), threading.main_thread().ident)"
    return execute_request_subshell_id(kc, code, subshell_id).split()


# Tests


def test_no_subshells():
    with new_kernel() as kc:
        # Test operation of separate channel thread without using any subshells.
        execute_request_subshell_id(kc, "a = 2*3", None)
        res = execute_request_subshell_id(kc, "print(a)", None)
        assert res == "6"


def test_supported():
    with new_kernel() as kc:
        msg_id = kc.kernel_info()
        reply = get_reply(kc, msg_id, TIMEOUT)
        assert "supported_features" in reply["content"]
        assert "kernel subshells" in reply["content"]["supported_features"]


def test_subshell_id_lifetime():
    with new_kernel() as kc:
        assert list_subshell_helper(kc)["subshell_id"] == []
        subshell_id = create_subshell_helper(kc)["subshell_id"]
        assert list_subshell_helper(kc)["subshell_id"] == [subshell_id]
        delete_subshell_helper(kc, subshell_id)
        assert list_subshell_helper(kc)["subshell_id"] == []


def test_thread_counts():
    with new_kernel() as kc:
        execute_request_subshell_id(kc, "import threading", None)
        nthreads = execute_thread_count(kc)

        subshell_id = create_subshell_helper(kc)["subshell_id"]
        nthreads2 = execute_thread_count(kc)
        assert nthreads2 > nthreads

        delete_subshell_helper(kc, subshell_id)
        nthreads3 = execute_thread_count(kc)
        assert nthreads3 == nthreads


def test_thread_ids():
    with new_kernel() as kc:
        execute_request_subshell_id(kc, "import threading", None)
        subshell_id = create_subshell_helper(kc)["subshell_id"]

        thread_id, main_thread_id = execute_thread_ids(kc)
        assert thread_id == main_thread_id

        thread_id, main_thread_id = execute_thread_ids(kc, subshell_id)  # This is the problem
        assert thread_id != main_thread_id

        delete_subshell_helper(kc, subshell_id)


@pytest.mark.parametrize("are_subshells", [(False, True), (True, False), (True, True)])
@pytest.mark.parametrize("overlap", [True, False])
def test_run_concurrently_sequence(are_subshells, overlap, request):
    if request.config.getvalue("--cov"):
        pytest.skip("Skip time-sensitive subshell tests if measuring coverage")

    with new_kernel() as kc:
        subshell_ids = [
            create_subshell_helper(kc)["subshell_id"] if is_subshell else None
            for is_subshell in are_subshells
        ]

        # Import time module before running time-sensitive subshell code
        # and use threading.Barrier to synchronise start of subshell code.
        execute_request_subshell_id(
            kc, "import threading as t, time; b=t.Barrier(2); print('ok')", None
        )

        sleep = 0.5
        if overlap:
            codes = [
                f"b.wait(); start0=True; end0=False; time.sleep({sleep}); end0=True",
                f"b.wait(); time.sleep({sleep / 2}); assert start0; assert not end0; time.sleep({sleep}); assert end0",
            ]
        else:
            codes = [
                f"b.wait(); start0=True; end0=False; time.sleep({sleep}); assert end1",
                f"b.wait(); time.sleep({sleep / 2}); assert start0; assert not end0; end1=True",
            ]

        msgs = []
        for subshell_id, code in zip(subshell_ids, codes, strict=False):
            msg = kc.session.msg("execute_request", {"code": code})
            msg["header"]["subshell_id"] = subshell_id
            kc.shell_channel.send(msg)
            msgs.append(msg)

        replies = get_replies(kc, [msg["msg_id"] for msg in msgs], timeout=None)

        for subshell_id in subshell_ids:
            if subshell_id:
                delete_subshell_helper(kc, subshell_id)

        for reply in replies:
            assert reply["content"]["status"] == "ok", reply


def test_create_while_execute():
    with new_kernel() as kc:
        # Send request to execute code on main subshell.
        msg = kc.session.msg("execute_request", {"code": "import time; time.sleep(0.05)"})
        kc.shell_channel.send(msg)

        # Create subshell via control channel.
        control_msg = kc.session.msg("create_subshell_request")
        kc.control_channel.send(control_msg)
        control_reply = get_reply(kc, control_msg["header"]["msg_id"], TIMEOUT, channel="control")
        subshell_id = control_reply["content"]["subshell_id"]
        control_date = control_reply["header"]["date"]

        # Get result message from main subshell.
        shell_date = get_reply(kc, msg["msg_id"])["header"]["date"]

        delete_subshell_helper(kc, subshell_id)

        assert control_date < shell_date


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="does not work on PyPy",
)
def test_shutdown_with_subshell():
    # Based on test_kernel.py::test_shutdown
    with new_kernel() as kc:
        km = kc.parent
        subshell_id = create_subshell_helper(kc)["subshell_id"]
        assert list_subshell_helper(kc)["subshell_id"] == [subshell_id]
        kc.shutdown()
        for _ in range(100):  # 10 s timeout
            if km.is_alive():
                time.sleep(0.1)
            else:
                break
        assert not km.is_alive()


@pytest.mark.parametrize("are_subshells", [(False, True), (True, False), (True, True)])
def test_execute_stop_on_error(are_subshells):
    # Based on test_message_spec.py::test_execute_stop_on_error, testing that exception
    # in one subshell aborts execution queue in that subshell but not others.
    with new_kernel() as kc:
        subshell_ids = [
            create_subshell_helper(kc)["subshell_id"] if is_subshell else None
            for is_subshell in are_subshells
        ]

        msg_ids = []

        msg = execute_request(
            kc, "import asyncio; await asyncio.sleep(1); raise ValueError()", subshell_ids[0]
        )
        msg_ids.append(msg["header"]["msg_id"])
        msg = execute_request(kc, "print('hello')", subshell_ids[0])
        msg_ids.append(msg["header"]["msg_id"])
        msg = execute_request(kc, "print('goodbye')", subshell_ids[0])
        msg_ids.append(msg["header"]["msg_id"])

        msg = execute_request(kc, "import time; time.sleep(1.5)", subshell_ids[1])
        msg_ids.append(msg["header"]["msg_id"])
        msg = execute_request(kc, "print('other')", subshell_ids[1])
        msg_ids.append(msg["header"]["msg_id"])

        replies = get_replies(kc, msg_ids)

        assert replies[0]["parent_header"]["subshell_id"] == subshell_ids[0]
        assert replies[1]["parent_header"]["subshell_id"] == subshell_ids[0]
        assert replies[2]["parent_header"]["subshell_id"] == subshell_ids[0]
        assert replies[3]["parent_header"]["subshell_id"] == subshell_ids[1]
        assert replies[4]["parent_header"]["subshell_id"] == subshell_ids[1]

        assert replies[0]["content"]["status"] == "error"
        assert replies[1]["content"]["status"] == "aborted"
        assert replies[2]["content"]["status"] == "aborted"
        assert replies[3]["content"]["status"] == "ok"
        assert replies[4]["content"]["status"] == "ok"

        # Check abort is cleared.
        msg = execute_request(kc, "print('check')", subshell_ids[0])
        reply = get_reply(kc, msg["msg_id"])
        assert reply["parent_header"]["subshell_id"] == subshell_ids[0]
        assert reply["content"]["status"] == "ok"

        # Cleanup
        for subshell_id in subshell_ids:
            if subshell_id:
                delete_subshell_helper(kc, subshell_id)


@pytest.mark.parametrize("are_subshells", [(False, True), (True, False), (True, True)])
def test_idle_message_parent_headers(are_subshells):
    with new_kernel() as kc:
        # import time module on main shell.
        msg = kc.session.msg("execute_request", {"code": "import time"})
        kc.shell_channel.send(msg)

        subshell_ids = [
            create_subshell_helper(kc)["subshell_id"] if is_subshell else None
            for is_subshell in are_subshells
        ]

        # Wait for all idle status messages to be received.
        for _ in range(1 + sum(are_subshells)):
            wait_for_idle(kc)

        msg_ids = []
        for subshell_id in subshell_ids:
            msg = execute_request(kc, "time.sleep(0.5)", subshell_id)
            msg_ids.append(msg["msg_id"])

        # Expect 4 status messages (2 busy, 2 idle) on iopub channel for the two execute_requests
        statuses = []
        timeout = TIMEOUT  # Combined timeout to receive all the status messages
        t0 = time.time()
        while True:
            status = kc.get_iopub_msg(timeout=timeout)
            if status["msg_type"] != "status" or status["parent_header"]["msg_id"] not in msg_ids:
                continue
            statuses.append(status)
            if len(statuses) == 4:
                break
            t1 = time.time()
            timeout -= t1 - t0
            t0 = t1

        execution_states = Counter(msg["content"]["execution_state"] for msg in statuses)
        assert execution_states["busy"] == 2
        assert execution_states["idle"] == 2

        parent_msg_ids = Counter(msg["parent_header"]["msg_id"] for msg in statuses)
        assert parent_msg_ids[msg_ids[0]] == 2
        assert parent_msg_ids[msg_ids[1]] == 2

        parent_subshell_ids = Counter(msg["parent_header"].get("subshell_id") for msg in statuses)
        assert parent_subshell_ids[subshell_ids[0]] == 2
        assert parent_subshell_ids[subshell_ids[1]] == 2

        # Cleanup
        for subshell_id in subshell_ids:
            if subshell_id:
                delete_subshell_helper(kc, subshell_id)


def test_silent_flag_in_subshells():
    """Verifies that the 'silent' flag suppresses output in main and subshell contexts."""
    with new_kernel() as kc:
        subshell_id = None
        try:
            flush_channels(kc)
            # Test silent execution in main shell
            msg_main_silent = execute_request(kc, "a=1", None, silent=True)
            reply_main_silent = get_reply(kc, msg_main_silent["header"]["msg_id"])
            assert reply_main_silent["content"]["status"] == "ok"

            # Test silent execution in subshell
            subshell_id = create_subshell_helper(kc)["subshell_id"]
            msg_sub_silent = execute_request(kc, "b=2", subshell_id, silent=True)
            reply_sub_silent = get_reply(kc, msg_sub_silent["header"]["msg_id"])
            assert reply_sub_silent["content"]["status"] == "ok"

            # Check for no iopub messages (other than status) from the silent requests
            for msg_id in [msg_main_silent["header"]["msg_id"], msg_sub_silent["header"]["msg_id"]]:
                while True:
                    try:
                        msg = kc.get_iopub_msg(timeout=0.2)
                        if msg["header"]["msg_type"] == "status":
                            continue
                        pytest.fail(
                            f"Silent execution produced unexpected IOPub message: {msg['header']['msg_type']}"
                        )
                    except Empty:
                        break

            # Test concurrent silent and non-silent execution
            msg_silent = execute_request(
                kc, "import time; time.sleep(0.5); c=3", subshell_id, silent=True
            )
            msg_noisy = execute_request(kc, "print('noisy')", None, silent=False)

            # Wait for both replies
            get_replies(kc, [msg_silent["header"]["msg_id"], msg_noisy["header"]["msg_id"]])

            # Verify that we only receive stream output from the noisy message
            stdout, stderr = assemble_output(
                kc.get_iopub_msg, parent_msg_id=msg_noisy["header"]["msg_id"]
            )
            assert "noisy" in stdout
            assert not stderr

            # Verify there is no output from the concurrent silent message
            while True:
                try:
                    msg = kc.get_iopub_msg(timeout=0.2)
                    if (
                        msg["header"]["msg_type"] == "status"
                        and msg["parent_header"].get("msg_id") == msg_silent["header"]["msg_id"]
                    ):
                        continue
                    if msg["parent_header"].get("msg_id") == msg_silent["header"]["msg_id"]:
                        pytest.fail(
                            "Silent execution in concurrent setting produced unexpected IOPub message"
                        )
                except Empty:
                    break
        finally:
            # Ensure subshell is always deleted
            if subshell_id:
                delete_subshell_helper(kc, subshell_id)



"""A thread for a shell channel."""

from __future__ import annotations

import asyncio
from threading import current_thread
from typing import Any

import zmq

from .subshell_manager import SubshellManager
from .thread import SHELL_CHANNEL_THREAD_NAME, BaseThread


class ShellChannelThread(BaseThread):
    """A thread for a shell channel.

    Communicates with shell/subshell threads via pairs of ZMQ inproc sockets.
    """

    def __init__(
        self,
        context: zmq.Context[Any],
        shell_socket: zmq.Socket[Any],
        **kwargs,
    ):
        """Initialize the thread."""
        super().__init__(name=SHELL_CHANNEL_THREAD_NAME, **kwargs)
        self._manager: SubshellManager | None = None
        self._zmq_context = context  # Avoid use of self._context
        self._shell_socket = shell_socket
        # Record the parent thread - the thread that started the app (usually the main thread)
        self.parent_thread = current_thread()

        self.asyncio_lock = asyncio.Lock()

    @property
    def manager(self) -> SubshellManager:
        # Lazy initialisation.
        if self._manager is None:
            assert current_thread() == self.parent_thread
            self._manager = SubshellManager(
                self._zmq_context,
                self.io_loop,
                self._shell_socket,
            )
        return self._manager

    def run(self) -> None:
        """Run the thread."""
        try:
            super().run()
        finally:
            if self._manager:
                self._manager.close()



"""test the IPython Kernel"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import ast
import os.path
import platform
import signal
import subprocess
import sys
import time
from datetime import datetime, timedelta
from subprocess import Popen
from tempfile import TemporaryDirectory

import IPython
import psutil
import pytest
from IPython.paths import locate_profile

from .utils import (
    TIMEOUT,
    assemble_output,
    execute,
    flush_channels,
    get_reply,
    kernel,
    new_kernel,
    wait_for_idle,
)


def _check_master(kc, expected=True, stream="stdout"):
    execute(kc=kc, code="import sys")
    flush_channels(kc)
    _msg_id, _content = execute(kc=kc, code="print(sys.%s._is_master_process())" % stream)
    stdout, _stderr = assemble_output(kc.get_iopub_msg)
    assert stdout.strip() == repr(expected)


def _check_status(content):
    """If status=error, show the traceback"""
    if content["status"] == "error":
        raise AssertionError("".join(["\n"] + content["traceback"]))


# printing tests


def test_simple_print():
    """simple print statement in kernel"""
    with kernel() as kc:
        _msg_id, _content = execute(kc=kc, code="print('hi')")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == "hi\n"
        assert stderr == ""
        _check_master(kc, expected=True)


def collect_outputs(get_iopub_msg, parent_msg_id, timeout=5):
    """Collect outputs until we get an idle message

    Returns list of complete output messages.
    """
    while True:
        msg = get_iopub_msg(timeout=timeout)
        msg_type = msg["msg_type"]
        content = msg["content"]

        if (
            msg["parent_header"]["msg_id"] == parent_msg_id
            and msg_type == "status"
            and content["execution_state"] == "idle"
        ):
            # idle message signals end of output
            break
        elif msg["msg_type"] in {"stream", "display_data"}:
            yield msg
        elif msg["msg_type"] == "error":
            tb = "\n".join(msg["content"]["traceback"])
            raise RuntimeError(f"Error during execution: {tb}")
        else:
            # other output, ignored
            print(msg["msg_type"])


@pytest.mark.parametrize("explicit_parent", [True, False])
def test_print_to_correct_cell_from_thread(explicit_parent: bool):
    """should print to the current cell unless

    get_ipython().set_parent sets the thread-local value,
    which supersedes the default.

    """
    code = f"""\
        from threading import Event, Thread
        from time import sleep
        from IPython.display import display

        explicit_parent = {explicit_parent}
        parent = get_ipython().get_parent()

        cell_start_event = Event()
        cell_end_event = Event()

        def thread_target():
            if explicit_parent:
                get_ipython().set_parent(parent)

            print("before", flush=True)
            display(1)
            cell_start_event.wait(timeout=10)
            cell_start_event.clear()

            print("during", flush=True)
            display(2)
            cell_end_event.set()
            cell_start_event.wait(timeout=10)
            cell_start_event.clear()
            print("after", flush=True)
            display(3)

        thread = Thread(target=thread_target)
        thread.start()
    """
    outputs = {}

    def add_output(msg):
        parent_id = msg["parent_header"]["msg_id"]
        if parent_id not in outputs:
            outputs[parent_id] = {
                "stdout": "",
                "stderr": "",
                "display_data": [],
            }
        cell_outputs = outputs[parent_id]
        msg_type = msg["header"]["msg_type"]
        content = msg["content"]
        if msg_type == "stream":
            cell_outputs[content["name"]] += content["text"]
        else:
            cell_outputs[msg_type].append(msg["content"]["data"]["text/plain"])

    with kernel() as kc:
        thread_msg_id = kc.execute(code)
        for msg in collect_outputs(kc.get_iopub_msg, thread_msg_id):
            add_output(msg)

        next_cell_msg_id = kc.execute("cell_start_event.set()\ncell_end_event.wait(timeout=10)")
        for msg in collect_outputs(kc.get_iopub_msg, next_cell_msg_id):
            add_output(msg)

        last_cell_msg_id = kc.execute("cell_start_event.set()\nthread.join()")
        for msg in collect_outputs(kc.get_iopub_msg, last_cell_msg_id):
            add_output(msg)
    print(outputs)
    if explicit_parent:
        # assert next_cell_msg_id not in outputs
        # assert last_cell_msg_id not in outputs
        thread_cell_output = outputs[thread_msg_id]
        assert thread_cell_output["stdout"] == "before\nduring\nafter\n"
        assert thread_cell_output["display_data"] == ["1", "2", "3"]
    else:
        thread_cell_output = outputs[thread_msg_id]
        assert thread_cell_output["stdout"] == "before\n"
        assert thread_cell_output["display_data"] == ["1"]
        next_cell_output = outputs[next_cell_msg_id]
        assert next_cell_output["stdout"] == "during\n"
        assert next_cell_output["display_data"] == ["2"]
        last_cell_output = outputs[last_cell_msg_id]
        assert last_cell_output["stdout"] == "after\n"
        assert last_cell_output["display_data"] == ["3"]


def test_print_to_correct_cell_from_child_thread():
    """should print to the cell that spawned the thread, not a subsequently run cell"""
    iterations = 5
    interval = 0.25
    code = f"""\
    from threading import Thread
    from time import sleep

    parent = get_ipython().get_parent()

    def child_target():
        get_ipython().set_parent(parent)
        for i in range({iterations}):
            print(i, end='', flush=True)
            sleep({interval})

    def parent_target():
        sleep({interval})
        thread = Thread(target=child_target)
        thread.start()
        thread.join()

    Thread(target=parent_target).start()
    """
    with kernel() as kc:
        thread_msg_id = kc.execute(code)
        _ = kc.execute("pass")

        received = 0
        while received < iterations:
            msg = kc.get_iopub_msg(timeout=interval * 2)
            if msg["msg_type"] != "stream":
                continue
            content = msg["content"]
            assert content["name"] == "stdout"
            assert content["text"] == str(received)
            # this is crucial as the parent header decides to which cell the output goes
            assert msg["parent_header"]["msg_id"] == thread_msg_id
            received += 1


def test_print_to_correct_cell_from_asyncio():
    """should print to the cell that scheduled the task, not a subsequently run cell"""
    iterations = 5
    interval = 0.25
    code = f"""\
    import asyncio

    async def async_task():
        for i in range({iterations}):
            print(i, end='', flush=True)
            await asyncio.sleep({interval})

    loop = asyncio.get_event_loop()
    loop.create_task(async_task());
    """
    with kernel() as kc:
        thread_msg_id = kc.execute(code)
        _ = kc.execute("pass")

        received = 0
        while received < iterations:
            msg = kc.get_iopub_msg(timeout=interval * 2)
            if msg["msg_type"] != "stream":
                continue
            content = msg["content"]
            assert content["name"] == "stdout"
            assert content["text"] == str(received)
            # this is crucial as the parent header decides to which cell the output goes
            assert msg["parent_header"]["msg_id"] == thread_msg_id
            received += 1


@pytest.mark.skip(reason="Currently don't capture during test as pytest does its own capturing")
def test_capture_fd():
    """simple print statement in kernel"""
    with kernel() as kc:
        iopub = kc.iopub_channel
        _msg_id, _content = execute(kc=kc, code="import os; os.system('echo capsys')")
        stdout, stderr = assemble_output(iopub)
        assert stdout == "capsys\n"
        assert stderr == ""
        _check_master(kc, expected=True)


@pytest.mark.skip(reason="Currently don't capture during test as pytest does its own capturing")
def test_subprocess_peek_at_stream_fileno():
    with kernel() as kc:
        iopub = kc.iopub_channel
        _msg_id, _content = execute(
            kc=kc,
            code="import subprocess, sys; subprocess.run(['python', '-c', 'import os; os.system(\"echo CAP1\"); print(\"CAP2\")'], stderr=sys.stderr)",
        )
        stdout, stderr = assemble_output(iopub)
        assert stdout == "CAP1\nCAP2\n"
        assert stderr == ""
        _check_master(kc, expected=True)


def test_sys_path():
    """test that sys.path doesn't get messed up by default"""
    with kernel() as kc:
        _msg_id, _content = execute(kc=kc, code="import sys; print(repr(sys.path))")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
    # for error-output on failure
    sys.stderr.write(stderr)

    sys_path = ast.literal_eval(stdout.strip())
    assert "" in sys_path


def test_sys_path_profile_dir():
    """test that sys.path doesn't get messed up when `--profile-dir` is specified"""

    with new_kernel(["--profile-dir", locate_profile("default")]) as kc:
        _msg_id, _content = execute(kc=kc, code="import sys; print(repr(sys.path))")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
    # for error-output on failure
    sys.stderr.write(stderr)

    sys_path = ast.literal_eval(stdout.strip())
    assert "" in sys_path


# the subprocess print tests fail in pytest,
# but manual tests in notebooks work fine...


@pytest.mark.flaky(max_runs=3)
@pytest.mark.skipif(
    sys.platform in {"win32", "darwin"} or sys.version_info >= (3, 14),
    reason="test doesn't reliably reproduce subprocess output capture",
)
def test_subprocess_print():
    """printing from forked mp.Process"""
    with new_kernel() as kc:
        _check_master(kc, expected=True)
        flush_channels(kc)
        np = 5
        code = "\n".join(
            [
                "import time",
                "import multiprocessing as mp",
                "pool = [mp.Process(target=print, args=('hello', i,)) for i in range(%i)]" % np,
                "for p in pool: p.start()",
                "for p in pool: p.join()",
                "time.sleep(0.5),",
            ]
        )

        _msg_id, _content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout.count("hello") == np, stdout
        for n in range(np):
            assert stdout.count(str(n)) == 1, stdout
        assert stderr == ""
        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


@pytest.mark.flaky(max_runs=3)
def test_subprocess_noprint():
    """mp.Process without print doesn't trigger iostream mp_mode"""
    with kernel() as kc:
        np = 5
        code = "\n".join(
            [
                "import multiprocessing as mp",
                "pool = [mp.Process(target=range, args=(i,)) for i in range(%i)]" % np,
                "for p in pool: p.start()",
                "for p in pool: p.join()",
            ]
        )

        _msg_id, _content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == ""
        assert stderr == ""

        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


@pytest.mark.flaky(max_runs=3)
@pytest.mark.skipif(
    sys.platform in {"win32", "darwin"} or sys.version_info >= (3, 14),
    reason="test doesn't reliably reproduce subprocess output capture",
)
def test_subprocess_error():
    """error in mp.Process doesn't crash"""
    with new_kernel() as kc:
        code = "\n".join(
            [
                "import multiprocessing as mp",
                "p = mp.Process(target=int, args=('hi',))",
                "p.start()",
                "p.join()",
            ]
        )

        _msg_id, _content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == ""
        assert "ValueError" in stderr

        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


# raw_input tests


def test_raw_input():
    """test input"""
    with kernel() as kc:
        input_f = "input"
        theprompt = "prompt> "
        code = f'print({input_f}("{theprompt}"))'
        kc.execute(code, allow_stdin=True)
        msg = kc.get_stdin_msg(timeout=TIMEOUT)
        assert msg["header"]["msg_type"] == "input_request"
        content = msg["content"]
        assert content["prompt"] == theprompt
        text = "some text"
        kc.input(text)
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        stdout, _stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == text + "\n"


def test_save_history():
    # Saving history from the kernel with %hist -f was failing because of
    # unicode problems on Python 2.
    with kernel() as kc, TemporaryDirectory() as td:
        file = os.path.join(td, "hist.out")
        execute("a=1", kc=kc)
        wait_for_idle(kc)
        execute('b="abcþ"', kc=kc)
        wait_for_idle(kc)
        _, reply = execute("%hist -f " + file, kc=kc)
        assert reply["status"] == "ok"
        with open(file, encoding="utf-8") as f:
            content = f.read()
        assert "a=1" in content
        assert 'b="abcþ"' in content


def test_smoke_faulthandler():
    pytest.importorskip("faulthandler", reason="this test needs faulthandler")
    with kernel() as kc:
        # Note: faulthandler.register is not available on windows.
        code = "\n".join(
            [
                "import sys",
                "import faulthandler",
                "import signal",
                "faulthandler.enable()",
                'if not sys.platform.startswith("win32"):',
                "    faulthandler.register(signal.SIGTERM)",
            ]
        )
        _, reply = execute(code, kc=kc)
        assert reply["status"] == "ok", reply.get("traceback", "")


def test_help_output():
    """ipython kernel --help-all works"""
    cmd = [sys.executable, "-m", "IPython", "kernel", "--help-all"]
    proc = subprocess.run(cmd, timeout=30, capture_output=True, check=True)
    assert proc.returncode == 0, proc.stderr
    assert b"Traceback" not in proc.stderr
    assert b"Options" in proc.stdout
    assert b"Class" in proc.stdout


def test_is_complete():
    with kernel() as kc:
        # There are more test cases for this in core - here we just check
        # that the kernel exposes the interface correctly.
        kc.is_complete("2+2")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "complete"

        # SyntaxError
        kc.is_complete("raise = 2")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "invalid"

        kc.is_complete("a = [1,\n2,")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "incomplete"
        assert reply["content"]["indent"] == ""

        # Cell magic ends on two blank lines for console UIs
        kc.is_complete("%%timeit\na\n\n")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "complete"


@pytest.mark.skipif(sys.platform != "win32", reason="only run on Windows")
def test_complete():
    with kernel() as kc:
        execute("a = 1", kc=kc)
        wait_for_idle(kc)
        cell = "import IPython\nb = a."
        kc.complete(cell)
        reply = kc.get_shell_msg(timeout=TIMEOUT)

    c = reply["content"]
    assert c["status"] == "ok"
    start = cell.find("a.")
    end = start + 2
    assert c["cursor_end"] == cell.find("a.") + 2
    assert c["cursor_start"] <= end

    # there are many right answers for cursor_start,
    # so verify application of the completion
    # rather than the value of cursor_start

    matches = c["matches"]
    assert matches
    for m in matches:
        completed = cell[: c["cursor_start"]] + m
        assert completed.startswith(cell)


def test_matplotlib_inline_on_import():
    pytest.importorskip("matplotlib", reason="this test requires matplotlib")
    with kernel() as kc:
        cell = "\n".join(
            ["import matplotlib, matplotlib.pyplot as plt", "backend = matplotlib.get_backend()"]
        )
        _, reply = execute(cell, user_expressions={"backend": "backend"}, kc=kc)
        _check_status(reply)
        backend_bundle = reply["user_expressions"]["backend"]
        _check_status(backend_bundle)
        assert "backend_inline" in backend_bundle["data"]["text/plain"]


def test_message_order():
    N = 100  # number of messages to test
    with kernel() as kc:
        _, reply = execute("a = 1", kc=kc)
        _check_status(reply)
        offset = reply["execution_count"] + 1
        cell = "a += 1\na"
        msg_ids = []
        # submit N executions as fast as we can
        for _ in range(N):
            msg_ids.append(kc.execute(cell))
        # check message-handling order
        for i, msg_id in enumerate(msg_ids, offset):
            reply = kc.get_shell_msg(timeout=TIMEOUT)
            _check_status(reply["content"])
            assert reply["content"]["execution_count"] == i
            assert reply["parent_header"]["msg_id"] == msg_id


@pytest.mark.skipif(
    sys.platform.startswith("linux") or sys.platform.startswith("darwin"),
    reason="test only on windows",
)
def test_unc_paths():
    with kernel() as kc, TemporaryDirectory() as td:
        drive_file_path = os.path.join(td, "unc.txt")
        with open(drive_file_path, "w+") as f:
            f.write("# UNC test")
        unc_root = "\\\\localhost\\C$"
        file_path = os.path.splitdrive(os.path.dirname(drive_file_path))[1]
        unc_file_path = os.path.join(unc_root, file_path[1:])

        kc.execute(f"cd {unc_file_path:s}")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        out, _err = assemble_output(kc.get_iopub_msg)
        assert unc_file_path in out

        flush_channels(kc)
        kc.execute(code="ls")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        out, _err = assemble_output(kc.get_iopub_msg)
        assert "unc.txt" in out

        kc.execute(code="cd")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="does not work on PyPy",
)
def test_shutdown():
    """Kernel exits after polite shutdown_request"""
    with new_kernel() as kc:
        km = kc.parent
        execute("a = 1", kc=kc)
        wait_for_idle(kc)
        kc.shutdown()
        for _ in range(300):  # 30s timeout
            if km.is_alive():
                time.sleep(0.1)
            else:
                break
        assert not km.is_alive()


def test_interrupt_during_input():
    """
    The kernel exits after being interrupted while waiting in input().

    input() appears to have issues other functions don't, and it needs to be
    interruptible in order for pdb to be interruptible.
    """
    with new_kernel() as kc:
        km = kc.parent
        msg_id = kc.execute("input()")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)


@pytest.mark.skipif(os.name == "nt", reason="Message based interrupt not supported on Windows")
def test_interrupt_with_message():
    with new_kernel() as kc:
        km = kc.parent
        km.kernel_spec.interrupt_mode = "message"
        msg_id = kc.execute("input()")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)


@pytest.mark.skipif(
    "__pypy__" in sys.builtin_module_names,
    reason="fails on pypy",
)
def test_interrupt_during_pdb_set_trace():
    """
    The kernel exits after being interrupted while waiting in pdb.set_trace().

    Merely testing input() isn't enough, pdb has its own issues that need
    to be handled in addition.

    This test will fail with versions of IPython < 7.14.0.
    """
    with new_kernel() as kc:
        km = kc.parent
        msg_id = kc.execute("import pdb; pdb.set_trace()")
        msg_id2 = kc.execute("3 + 4")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)
        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id2, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id2)


def test_control_thread_priority():
    N = 5
    with new_kernel() as kc:
        msg_id = kc.execute("pass")
        get_reply(kc, msg_id)

        sleep_msg_id = kc.execute("import asyncio; await asyncio.sleep(2)")

        # submit N shell messages
        shell_msg_ids = []
        for i in range(N):
            shell_msg_ids.append(kc.execute(f"i = {i}"))

        # ensure all shell messages have arrived at the kernel before any control messages
        time.sleep(0.5)
        # at this point, shell messages should be waiting in msg_queue,
        # rather than zmq while the kernel is still in the middle of processing
        # the first execution

        # now send N control messages
        control_msg_ids = []
        for _ in range(N):
            msg = kc.session.msg("kernel_info_request", {})
            kc.control_channel.send(msg)
            control_msg_ids.append(msg["header"]["msg_id"])

        # finally, collect the replies on both channels for comparison
        get_reply(kc, sleep_msg_id)
        shell_replies = []
        for msg_id in shell_msg_ids:
            shell_replies.append(get_reply(kc, msg_id))

        control_replies = []
        for msg_id in control_msg_ids:
            control_replies.append(get_reply(kc, msg_id, channel="control"))

    # verify that all control messages were handled before all shell messages
    shell_dates = [msg["header"]["date"] for msg in shell_replies]
    control_dates = [msg["header"]["date"] for msg in control_replies]
    # comparing first to last ought to be enough, since queues preserve order
    # use <= in case of very-fast handling and/or low resolution timers
    assert control_dates[-1] <= shell_dates[0]


def test_sequential_control_messages():
    with new_kernel() as kc:
        msg_id = kc.execute("import time")
        get_reply(kc, msg_id)

        # Send multiple messages on the control channel.
        # Using execute messages to vary duration.
        sleeps = [0.6, 0.3, 0.1]

        # Prepare messages
        msgs = [
            kc.session.msg("execute_request", {"code": f"time.sleep({sleep})"}) for sleep in sleeps
        ]
        msg_ids = [msg["header"]["msg_id"] for msg in msgs]

        # Submit messages
        for msg in msgs:
            kc.control_channel.send(msg)

        # Get replies
        replies = [get_reply(kc, msg_id, channel="control") for msg_id in msg_ids]

        def ensure_datetime(arg):
            # Support arg which is a datetime or str.
            if isinstance(arg, str):
                if sys.version_info[:2] < (3, 11) and arg.endswith("Z"):
                    # Python < 3.11 doesn't support "Z" suffix in datetime.fromisoformat,
                    # so use alternative timezone format.
                    # https://github.com/python/cpython/issues/80010
                    arg = arg[:-1] + "+00:00"
                return datetime.fromisoformat(arg)
            return arg

        # Check messages are processed in order, one at a time, and of a sensible duration.
        previous_end = None
        for reply, sleep in zip(replies, sleeps, strict=False):
            start = ensure_datetime(reply["metadata"]["started"])
            end = ensure_datetime(reply["header"]["date"])

            if previous_end is not None:
                assert start >= previous_end
            previous_end = end

            assert end >= start + timedelta(seconds=sleep)


def _child():
    print("in child", os.getpid())

    def _print_and_exit(sig, frame):
        print(f"Received signal {sig}")
        # take some time so retries are triggered
        time.sleep(0.5)
        sys.exit(-sig)

    signal.signal(signal.SIGTERM, _print_and_exit)
    time.sleep(30)


def _start_children():
    ip = IPython.get_ipython()  # type:ignore[attr-defined]
    ns = ip.user_ns

    cmd = [sys.executable, "-c", f"from {__name__} import _child; _child()"]
    child_pg = Popen(cmd, start_new_session=False)
    child_newpg = Popen(cmd, start_new_session=True)
    ns["pid"] = os.getpid()
    ns["child_pg"] = child_pg.pid
    ns["child_newpg"] = child_newpg.pid
    # give them time to start up and register signal handlers
    time.sleep(1)


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="does not work on PyPy",
)
@pytest.mark.skipif(
    sys.platform.lower() == "linux",
    reason="Stalls on linux",
)
def test_shutdown_subprocesses():
    """Kernel exits after polite shutdown_request"""
    with new_kernel() as kc:
        km = kc.parent
        _msg_id, reply = execute(
            f"from {__name__} import _start_children\n_start_children()",
            kc=kc,
            user_expressions={
                "pid": "pid",
                "child_pg": "child_pg",
                "child_newpg": "child_newpg",
            },
        )
        print(reply)
        expressions = reply["user_expressions"]
        kernel_process = psutil.Process(int(expressions["pid"]["data"]["text/plain"]))
        child_pg = psutil.Process(int(expressions["child_pg"]["data"]["text/plain"]))
        child_newpg = psutil.Process(int(expressions["child_newpg"]["data"]["text/plain"]))
        wait_for_idle(kc)

        kc.shutdown()
        for _ in range(300):  # 30s timeout
            if km.is_alive():
                time.sleep(0.1)
            else:
                break
        assert not km.is_alive()
        assert not kernel_process.is_running()
        # child in the process group shut down
        assert not child_pg.is_running()
        # child outside the process group was not shut down (unix only)
        if os.name != "nt":
            assert child_newpg.is_running()
        try:
            child_newpg.terminate()
        except psutil.NoSuchProcess:
            pass


def test_parent_header_and_ident():
    # Kernel._parent_ident is private but kept for backward compatibility,
    # see https://github.com/jupyterlab/jupyterlab/issues/17785
    with kernel() as kc:
        # get_parent('shell')
        msg_id, _ = execute(
            kc=kc,
            code="k=get_ipython().kernel; p=k.get_parent('shell'); print(p['header']['msg_id'], p['header']['session'])",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        check_msg_id, session = stdout.split()
        assert check_msg_id == msg_id
        assert check_msg_id.startswith(msg_id)

        # _parent_ident['shell']
        msg_id, _ = execute(kc=kc, code="print(k._parent_ident['shell'])")
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        assert stdout == f"[b'{session}']\n"

        # Send a control message
        msg = kc.session.msg("kernel_info_request")
        kc.control_channel.send(msg)
        control_msg_id = msg["header"]["msg_id"]
        assemble_output(kc.get_iopub_msg, parent_msg_id=control_msg_id)

        # get_parent('control')
        msg_id, _ = execute(
            kc=kc,
            code="p=k.get_parent('control'); print(p['header']['msg_id'], p['header']['session'])",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        check_msg_id, session = stdout.split()
        assert check_msg_id == control_msg_id
        assert check_msg_id.startswith(control_msg_id)

        # _parent_ident['control']
        msg_id, _ = execute(kc=kc, code="print(k._parent_ident['control'])")
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        assert stdout == f"[b'{session}']\n"


def test_context_vars():
    with new_kernel() as kc:
        msg_id, _ = execute(
            kc=kc,
            code="from contextvars import ContextVar, copy_context\nctxvar = ContextVar('var', default='default')",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)

        msg_id, _ = execute(
            kc=kc,
            code="print(ctxvar.get())",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        assert stdout.strip() == "default"

        msg_id, _ = execute(
            kc=kc,
            code="ctxvar.set('set'); print(ctxvar.get())",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        assert stdout.strip() == "set"

        msg_id, _ = execute(
            kc=kc,
            code="print(ctxvar.get())",
        )
        stdout, _ = assemble_output(kc.get_iopub_msg, parent_msg_id=msg_id)
        assert stdout.strip() == "set"



"""Base class for threads."""

from threading import Thread

from tornado.ioloop import IOLoop

CONTROL_THREAD_NAME = "Control"
SHELL_CHANNEL_THREAD_NAME = "Shell channel"


class BaseThread(Thread):
    """Base class for threads."""

    def __init__(self, **kwargs):
        """Initialize the thread."""
        super().__init__(**kwargs)
        self.io_loop = IOLoop(make_current=False)
        self.pydev_do_not_trace = True
        self.is_pydev_daemon_thread = True

    def run(self) -> None:
        """Run the thread."""
        try:
            self.io_loop.start()
        finally:
            self.io_loop.close()

    def stop(self) -> None:
        """Stop the thread.

        This method is threadsafe.
        """
        self.io_loop.add_callback(self.io_loop.stop)



"""An in-process qt console app."""

import os

import tornado
from IPython.lib import guisupport
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.rich_ipython_widget import RichIPythonWidget

assert tornado.version_info >= (6, 1)


def print_process_id():
    """Print the process id."""
    print("Process ID is:", os.getpid())


def main():
    """The main entry point."""
    # Print the ID of the main process
    print_process_id()

    app = guisupport.get_app_qt4()

    # Create an in-process kernel
    # >>> print_process_id()
    # will print the same process ID as the main process
    kernel_manager = QtInProcessKernelManager()
    kernel_manager.start_kernel()
    kernel = kernel_manager.kernel
    kernel.gui = "qt4"
    kernel.shell.push({"foo": 43, "print_process_id": print_process_id})

    kernel_client = kernel_manager.client()
    kernel_client.start_channels()

    def stop():
        kernel_client.stop_channels()
        kernel_manager.shutdown_kernel()
        app.exit()

    control = RichIPythonWidget()
    control.kernel_manager = kernel_manager
    control.kernel_client = kernel_client
    control.exit_requested.connect(stop)
    control.show()

    guisupport.start_event_loop_qt4(app)


if __name__ == "__main__":
    main()



"""A thread for a control channel."""

from .thread import CONTROL_THREAD_NAME, BaseThread


class ControlThread(BaseThread):
    """A thread for a control channel."""

    def __init__(self, **kwargs):
        """Initialize the thread."""
        super().__init__(name=CONTROL_THREAD_NAME, **kwargs)



"""Tests for zmq shell / display publisher."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import unittest
import warnings
from queue import Queue
from threading import Thread
from unittest.mock import MagicMock

import pytest
import zmq
from jupyter_client.session import Session
from traitlets import Int

from ipykernel.zmqshell import (  # type:ignore
    InteractiveShell,
    KernelMagics,
    ZMQDisplayPublisher,
    ZMQInteractiveShell,
)

try:
    from IPython.core.history import HistoryOutput
except ImportError:
    HistoryOutput = None  # type: ignore[assignment,misc]


class NoReturnDisplayHook:
    """
    A dummy DisplayHook which allows us to monitor
    the number of times an object is called, but which
    does *not* return a message when it is called.
    """

    call_count = 0

    def __call__(self, obj):
        self.call_count += 1


class ReturnDisplayHook(NoReturnDisplayHook):
    """
    A dummy DisplayHook with the same counting ability
    as its base class, but which also returns the same
    message when it is called.
    """

    def __call__(self, obj):
        super().__call__(obj)
        return obj


class CounterSession(Session):
    """
    This is a simple subclass to allow us to count
    the calls made to the session object by the display
    publisher.
    """

    send_count = Int(0)

    def send(self, *args, **kwargs):
        """
        A trivial override to just augment the existing call
        with an increment to the send counter.
        """
        self.send_count += 1
        super().send(*args, **kwargs)


class ZMQDisplayPublisherTests(unittest.TestCase):
    """
    Tests the ZMQDisplayPublisher in zmqshell.py
    """

    def setUp(self):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PUB)
        self.session = CounterSession()

        self.disp_pub = ZMQDisplayPublisher(session=self.session, pub_socket=self.socket)

    def tearDown(self):
        """
        We need to close the socket in order to proceed with the
        tests.
        TODO - There is still an open file handler to '/dev/null',
        presumably created by zmq.
        """
        self.disp_pub.clear_output()
        self.socket.close()
        self.context.term()

    def test_display_publisher_creation(self):
        """
        Since there's no explicit constructor, here we confirm
        that keyword args get assigned correctly, and override
        the defaults.
        """
        assert self.disp_pub.session == self.session
        assert self.disp_pub.pub_socket == self.socket

    def test_thread_local_hooks(self):
        """
        Confirms that the thread_local attribute is correctly
        initialised with an empty list for the display hooks
        """
        assert self.disp_pub._hooks == []

        def hook(msg):
            return msg

        self.disp_pub.register_hook(hook)
        assert self.disp_pub._hooks == [hook]

        q: Queue = Queue()

        def set_thread_hooks():
            q.put(self.disp_pub._hooks)

        t = Thread(target=set_thread_hooks)
        t.start()
        thread_hooks = q.get(timeout=10)
        assert thread_hooks == []

    def test_publish(self):
        """
        Publish should prepare the message and eventually call
        `send` by default.
        """
        data = dict(a=1)
        assert self.session.send_count == 0
        self.disp_pub.publish(data)
        assert self.session.send_count == 1

    def test_display_hook_halts_send(self):
        """
        If a hook is installed, and on calling the object
        it does *not* return a message, then we assume that
        the message has been consumed, and should not be
        processed (`sent`) in the normal manner.
        """
        data = dict(a=1)
        hook = NoReturnDisplayHook()

        self.disp_pub.register_hook(hook)
        assert hook.call_count == 0
        assert self.session.send_count == 0

        self.disp_pub.publish(data)

        assert hook.call_count == 1
        assert self.session.send_count == 0

    def test_display_hook_return_calls_send(self):
        """
        If a hook is installed and on calling the object
        it returns a new message, then we assume that this
        is just a message transformation, and the message
        should be sent in the usual manner.
        """
        data = dict(a=1)
        hook = ReturnDisplayHook()

        self.disp_pub.register_hook(hook)
        assert hook.call_count == 0
        assert self.session.send_count == 0

        self.disp_pub.publish(data)

        assert hook.call_count == 1
        assert self.session.send_count == 1

    def test_unregister_hook(self):
        """
        Once a hook is unregistered, it should not be called
        during `publish`.
        """
        data = dict(a=1)
        hook = NoReturnDisplayHook()

        self.disp_pub.register_hook(hook)
        assert hook.call_count == 0
        assert self.session.send_count == 0

        self.disp_pub.publish(data)

        assert hook.call_count == 1
        assert self.session.send_count == 0

        #
        # After unregistering the `NoReturn` hook, any calls
        # to publish should *not* got through the DisplayHook,
        # but should instead hit the usual `session.send` call
        # at the end.
        #
        # As a result, the hook call count should *not* increase,
        # but the session send count *should* increase.
        #
        first = self.disp_pub.unregister_hook(hook)
        self.disp_pub.publish(data)

        assert bool(first)
        assert hook.call_count == 1
        assert self.session.send_count == 1

        #
        # If a hook is not installed, `unregister_hook`
        # should return false.
        #
        second = self.disp_pub.unregister_hook(hook)
        assert not bool(second)

    @unittest.skipIf(HistoryOutput is None, "HistoryOutput not available")
    def test_display_stored_in_history(self):
        """
        Test that published display data gets stored in shell history
        for %notebook magic support, and not stored when disabled.
        """
        for enable in [False, True]:
            # Mock shell with history manager
            mock_shell = MagicMock()
            mock_shell.execution_count = 1
            mock_shell.history_manager.outputs = dict()
            mock_shell.display_pub._in_post_execute = False

            self.disp_pub.shell = mock_shell
            self.disp_pub.store_display_history = enable

            data = {"text/plain": "test output"}
            self.disp_pub.publish(data)

            if enable:
                # Check that output was stored in history
                stored_outputs = mock_shell.history_manager.outputs[1]
                assert len(stored_outputs) == 1
                assert stored_outputs[0].output_type == "display_data"
                assert stored_outputs[0].bundle == data
            else:
                # Should not store anything in history
                assert mock_shell.history_manager.outputs == {}


def test_magics(tmp_path):
    context = zmq.Context()
    socket = context.socket(zmq.PUB)
    shell = InteractiveShell()
    shell.user_ns["hi"] = 1
    magics = KernelMagics(shell)

    tmp_file = tmp_path / "test.txt"
    tmp_file.write_text("hi", "utf8")
    magics.edit(str(tmp_file))
    payload = shell.payload_manager.read_payload()[0]
    assert payload["filename"] == str(tmp_file)

    magics.clear([])
    magics.less(str(tmp_file))
    if os.name == "posix":
        magics.man("ls")
    magics.autosave("10")

    socket.close()
    context.destroy()


def test_zmq_interactive_shell(kernel):
    shell = ZMQInteractiveShell()

    with pytest.raises(RuntimeError):
        shell.enable_gui("tk")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        shell.data_pub_class = MagicMock()  # type:ignore
        shell.data_pub
    shell.kernel = kernel
    shell.set_next_input("hi")
    assert shell.get_parent() == {}
    if os.name == "posix":
        shell.system_piped("ls")
    else:
        shell.system_piped("dir")
    shell.ask_exit()


if __name__ == "__main__":
    unittest.main()
